from ngsolve import *
from netgen.geom2d import SplineGeometry
from netgen.geom2d import unit_square

#******************************************************************************************
# Domain + Mesh

def topBottom():
    geometry = SplineGeometry()
    pnts     = [ (0, 0), (1, 0),  (1, 0.5), (1, 1), (0, 1), (0, 0.5)]
    pnums    = [geometry.AppendPoint(*p) for p in pnts]
    # start-point, end-point, boundary-condition, domain on left side, domain on right side:
    lines = [ (pnums[0], pnums[1], "gamma2",  2, 0),
              (pnums[1], pnums[2], "gamma2",  2, 0), 
              (pnums[2], pnums[3], "gamma1",  1, 0),
              (pnums[3], pnums[4], "gamma1",  1, 0),
              (pnums[4], pnums[5], "gamma1",  1, 0),
              (pnums[5], pnums[0], "gamma2",  2, 0),
              (pnums[5], pnums[2], "gamma12", 1, 2)]
    for p1,p2,bc,left,right in lines:
        geometry.Append( ["line", p1, p2], bc=bc, leftdomain=left, rightdomain=right)

    geometry.SetMaterial(1, "domain1")
    geometry.SetMaterial(2, "domain2")
        
    return geometry

def GetInterfaceIndicator(mesh):
    dom_indicator = GridFunction( L2(mesh, order=0) )
    dom_indicator.Set( CoefficientFunction([1,2]) )
  
    ffes            = FacetFESpace(mesh, order=0)
    dom_ind_average = GridFunction(ffes)
    mf  = BilinearForm(ffes, symmetric=True)
    mf += SymbolicBFI(ffes.TrialFunction()*ffes.TestFunction(), element_boundary=True)
    mf.Assemble()
    ff  = LinearForm(ffes)
    ff += SymbolicLFI(dom_indicator*ffes.TestFunction(), element_boundary=True)
    ff.Assemble()
    dom_ind_average.vec.data = mf.mat.Inverse() * ff.vec

    interface_indicator = BitArray(ffes.ndof)
    for i in range(len(dom_ind_average.vec)):
        if (dom_ind_average.vec[i] > 1.75) or (dom_ind_average.vec[i] < 1.25):
            interface_indicator[i] = False
        else:
            interface_indicator[i] = True
    return interface_indicator

mesh = Mesh(topBottom().GenerateMesh (maxh=1.0/32.0))

#******************************************************************************************
# HDG for advection equation

tau = 1e-3
theta = 1.0
tend = 1.0

XCell = L2(mesh, order=1, dgjumps=True )
XFace = FacetFESpace(mesh, order=1)
XC    = FESpace([XCell, XFace])

c,cbar = XC.TrialFunction()
d,dbar = XC.TestFunction()

b  = CoefficientFunction((0, -1))
bn = b*specialcf.normal(2)

xbnd = CoefficientFunction(x*(1-x))
ybnd = CoefficientFunction(y*(1-y))

ubnd = CoefficientFunction(1.0) # Boundary condition

ab = BilinearForm(XC)
ab_cells  = c*d/tau-c*(b*grad(d))
ab_facets = bn*(c-IfPos(bn,0,1)*(c-cbar))*(d-dbar) \
            + (IfPos(xbnd,0,1)+IfPos(ybnd,0,1))*(1-IfPos(bn,0,1))*bn*cbar*dbar
ab += SymbolicBFI(ab_cells)
ab += SymbolicBFI(ab_facets, element_boundary=True)
ab.Assemble()
inv = ab.mat.Inverse(XC.FreeDofs())

fb = LinearForm(XC)   
fb.Assemble()
res = fb.vec.CreateVector()

gfb = GridFunction(XC)
gfb.components[0].Set(1.0+0.0*x)
Draw (gfb.components[0], autoscale=True, sd=2)

rhs = BilinearForm(XC, nonassemble=True)
rhs += SymbolicBFI( c*d/tau )
rhs += SymbolicBFI(-(IfPos(xbnd,0,1)+IfPos(ybnd,0,1))*IfPos(bn,0,1)*bn*ubnd*dbar, element_boundary=True)

t = 0
with TaskManager():
    while t < tend:
        rhs.Apply(gfb.vec, res)
        gfb.vec.data = inv * res        
        t += tau
        print("t=",t)
        Redraw()

err_c = sqrt(Integrate((gfb.components[0]-1.0)**2, mesh))
print("error c: ", err_c)


